/*
 * (c) 2008-2009 Adam Lackorzynski <adam@os.inf.tu-dresden.de>
 *     economic rights: Technische Universität Dresden (Germany)
 * Copyright (C) 2017, 2018, 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
/*
 * Access the I2C peripherals on the MIPS Creator CI20 board.
 */

#include <l4/devices/cpm-jz4780.h>
#include <l4/devices/gpio-jz4780.h>
#include <l4/devices/i2c-jz4780.h>
#include <l4/devices/memory.h>
#include <l4/io/io.h>
#include <l4/libedid/edid.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/vbus/vbus.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>



enum {
  PMSDA  = 30,  /* via PORTD */
  PMSCL  = 31,  /* via PORTD */
  PWM3   = 3,   /* via PORTE */
  PWM4   = 4,   /* via PORTE */
  RTCSDA = 12,  /* via PORTE */
  RTCSCL = 13,  /* via PORTE */
  DDCSCL = 24,  /* via PORTF */
  DDCSDA = 25,  /* via PORTF */
};



/* Device and resource discovery. */

static long item_in_range(long start, long end, long index)
{
  if (start < end)
    return start + index;
  else
    return start - index;
}

static void show_deviceid(uint8_t buf[])
{
  printf("Manufacturer %x part %x revision %x\n",
         (buf[0] << 4) | (buf[1] >> 4),
         ((buf[1] & 0xf) << 5) | (buf[2] >> 3),
         buf[2] & 0x7);
}

static void show_data(uint8_t buf[], unsigned pos)
{
  unsigned i;

  if (pos)
  {
    printf("Read %d bytes from bus.\n", pos);
    for (i = 0; i < pos; i++)
      printf(" %02x", buf[i]);
    printf("\n");
  }
  else
    printf("No reply from bus.\n");
}

static long i2c_read(void *i2c_channel, uint8_t *buf, unsigned length,
                     l4_cap_idx_t irqcap)
{
  l4_msgtag_t tag;
  long err;

  jz4780_i2c_start_read(i2c_channel, buf, length);

  while (!jz4780_i2c_read_done(i2c_channel))
  {
    tag = l4_irq_receive(irqcap, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4_timeout_rel(1, 20)));

    if ((err = l4_ipc_error(tag, l4_utcb())))
      return err;

    if (jz4780_i2c_failed(i2c_channel))
      return 0;

    jz4780_i2c_read(i2c_channel);
  }

  return jz4780_i2c_have_read(i2c_channel);
}

static long i2c_write(void *i2c_channel, uint8_t *buf, unsigned length,
                      l4_cap_idx_t irqcap)
{
  l4_msgtag_t tag;
  long err;

  jz4780_i2c_start_write(i2c_channel, buf, length);

  while (!jz4780_i2c_write_done(i2c_channel))
  {
    tag = l4_irq_receive(irqcap, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4_timeout_rel(1, 20)));

    if ((err = l4_ipc_error(tag, l4_utcb())))
      return err;

    if (jz4780_i2c_failed(i2c_channel))
      return 0;

    jz4780_i2c_write(i2c_channel);
  }

  return jz4780_i2c_have_written(i2c_channel);
}

static void i2c_scan(void *i2c_channel, l4_cap_idx_t irqcap)
{
  uint8_t buf[1];
  unsigned int address;

  for (address = 0; address < 0x20; address++)
    printf("%02x ", address);
  printf("\n");

  for (address = 0; address < 0x20; address++)
    printf("-- ");

  for (address = 0; address < 0x80; address++)
  {
    if ((address % 32) == 0)
      printf("\n");

    jz4780_i2c_set_target(i2c_channel, address);
    buf[0] = 0;
    i2c_write(i2c_channel, buf, 1, irqcap);
    i2c_read(i2c_channel, buf, 1, irqcap);
    jz4780_i2c_stop(i2c_channel);

    if (!jz4780_i2c_failed(i2c_channel))
      printf("%02x ", address);
    else
      printf("?? ");
  }

  printf("\n");
  for (address = 0; address < 0x20; address++)
    printf("-- ");
  printf("\n\n");
}

static void ddc_read(void *i2c_channel, l4_cap_idx_t irqcap)
{
  /* Buffer for reading. */

  uint8_t buf[128];
  long pos;

  /* DDC EDID details. */

  unsigned width = 0, height = 0;

  printf("Trying DDC on I2C4...\n");

  /* Attempt to read from address 0x50 for DDC.
     See: drivers/video/fbdev/core/fb_ddc.c */

  jz4780_i2c_set_target(i2c_channel, 0x50);
  buf[0] = 0;
  i2c_write(i2c_channel, buf, 1, irqcap);
  memset(buf, 0, 128);
  pos = i2c_read(i2c_channel, buf, 128, irqcap);
  jz4780_i2c_stop(i2c_channel);

  if (pos <= 0)
  {
    printf("Read failed.\n");
    return;
  }

  show_data(buf, pos);

  /* Attempt to decode EDID information. */

  libedid_prefered_resolution(buf, &width, &height);
  printf("Preferred resolution: %d x %d\n", width, height);

  libedid_dump_standard_timings(buf);
}

static void pmic_out3_control(void *i2c_channel, l4_cap_idx_t irqcap)
{
  /* Buffer for communicating. */

  uint8_t buf[2];

  /* Attempt to read from address 0x5a for PMIC OUT3. */

  jz4780_i2c_set_target(i2c_channel, 0x5a);
  buf[0] = 0x32;
  i2c_write(i2c_channel, buf, 1, irqcap);
  i2c_read(i2c_channel, buf, 1, irqcap);
  jz4780_i2c_stop(i2c_channel);

  if (!jz4780_i2c_failed(i2c_channel))
    printf("Read: %02x\n", buf[0]);
  else
    printf("Read failed.\n");

#if 0
  /* Disable the regulator. */

  printf("Updating...\n");

  buf[1] = buf[0] & ~0x80;
  buf[0] = 0x32;
  i2c_write(i2c_channel, buf, 2, irqcap);
  jz4780_i2c_stop(i2c_channel);

  /* Read back from the register. Seemed to give 0xff, which makes no real
     sense, although the regulator did bring the voltage level low. */

  buf[0] = 0x32;
  i2c_write(i2c_channel, buf, 1, irqcap);
  i2c_read(i2c_channel, buf, 1, irqcap);
  jz4780_i2c_stop(i2c_channel);

  if (!jz4780_i2c_failed(i2c_channel))
    printf("Read: %02x\n", buf[0]);
  else
    printf("Read failed.\n");
#endif
}

int main(void)
{
  long err;

  /* Peripheral memory. */

  l4_addr_t gpio_base = 0, gpio_base_end = 0;
  l4_addr_t i2c_base = 0, i2c_base_end = 0;
  l4_addr_t cpm_base = 0, cpm_base_end = 0;
  l4_addr_t port_d, port_d_end;
  l4_addr_t port_e, port_e_end;
  l4_addr_t port_f, port_f_end;

  /* Peripheral abstractions. */

  void *gpio_port_d, *gpio_port_e, *gpio_port_f;
  void *i2c, *cpm;
  void *i2c0, *i2c4;

  /* IRQ resources. */

  l4_uint32_t i2c_irq_start = 0, i2c_irq_end = 0;
  l4_cap_idx_t icucap, irq0cap, irq4cap;

  /* Obtain capabilities for the interrupt controller and an interrupt. */

  irq0cap = l4re_util_cap_alloc();
  irq4cap = l4re_util_cap_alloc();
  icucap = l4re_env_get_cap("icu");

  if (l4_is_invalid_cap(icucap))
  {
    printf("No 'icu' capability available in the virtual bus.\n");
    return 1;
  }

  if (l4_is_invalid_cap(irq0cap) || l4_is_invalid_cap(irq4cap))
  {
    printf("Capabilities not available for interrupts.\n");
    return 1;
  }

  /* Obtain resource details describing the interrupt for I2C channel 4. */

  printf("Access IRQ...\n");

  if (get_irq("jz4780-i2c", &i2c_irq_start, &i2c_irq_end) < 0)
    return 1;

  printf("IRQ range at %d...%d.\n", i2c_irq_start, i2c_irq_end);

  /* Obtain resource details describing I/O memory. */

  printf("Access GPIO...\n");

  if (get_memory("jz4780-gpio", &gpio_base, &gpio_base_end) < 0)
    return 1;

  printf("GPIO at 0x%lx...0x%lx.\n", gpio_base, gpio_base_end);

  printf("Access CPM...\n");

  if (get_memory("jz4780-cpm", &cpm_base, &cpm_base_end) < 0)
    return 1;

  printf("CPM at 0x%lx...0x%lx.\n", cpm_base, cpm_base_end);

  printf("Access I2C...\n");

  if (get_memory("jz4780-i2c", &i2c_base, &i2c_base_end) < 0)
    return 1;

  printf("I2C at 0x%lx...0x%lx.\n", i2c_base, i2c_base_end);

  /* Create interrupt objects. */

  err = l4_error(l4_factory_create_irq(l4re_global_env->factory, irq0cap)) ||
        l4_error(l4_factory_create_irq(l4re_global_env->factory, irq4cap));

  if (err)
  {
    printf("Could not create IRQ object: %lx\n", err);
    return 1;
  }

  /* Bind interrupt objects to IRQ numbers. */

  err = l4_error(l4_icu_bind(icucap,
                             item_in_range(i2c_irq_start, i2c_irq_end, 0),
                             irq0cap)) ||
        l4_error(l4_icu_bind(icucap,
                             item_in_range(i2c_irq_start, i2c_irq_end, 4),
                             irq4cap));

  if (err)
  {
    printf("Could not bind IRQ to the ICU: %ld\n", err);
    return 1;
  }

  /* Attach ourselves to the interrupt handler. */

  err = l4_error(l4_rcv_ep_bind_thread(irq0cap, l4re_env()->main_thread, 0)) ||
        l4_error(l4_rcv_ep_bind_thread(irq4cap, l4re_env()->main_thread, 4));

  if (err)
  {
    printf("Could not attach to IRQs: %ld\n", err);
    return 1;
  }

  /* Configure pins. */

  port_d = gpio_base + 0x300;
  port_d_end = port_d + 0x100;
  port_e = gpio_base + 0x400;
  port_e_end = port_e + 0x100;
  port_f = gpio_base + 0x500;
  port_f_end = port_f + 0x100;

  printf("PORTD at 0x%lx...0x%lx.\n", port_d, port_d_end);
  printf("PORTE at 0x%lx...0x%lx.\n", port_e, port_e_end);
  printf("PORTF at 0x%lx...0x%lx.\n", port_f, port_f_end);

  gpio_port_d = jz4780_gpio_init(port_d, port_d_end, 32, 0xffff4fff, 0x0000b000);
  gpio_port_e = jz4780_gpio_init(port_e, port_e_end, 32, 0xfffff37c, 0x00000483);
  gpio_port_f = jz4780_gpio_init(port_f, port_f_end, 32, 0x7fa7f00f, 0x00580ff0);

  printf("Set up GPIO pins...\n");

  jz4780_gpio_config_pad(gpio_port_d, PMSCL, Function_alt, 0);
  jz4780_gpio_config_pad(gpio_port_d, PMSDA, Function_alt, 0);

  jz4780_gpio_config_pad(gpio_port_e, RTCSCL, Function_alt, 1);
  jz4780_gpio_config_pad(gpio_port_e, RTCSDA, Function_alt, 1);

  jz4780_gpio_config_pad(gpio_port_f, DDCSCL, Function_alt, 1);
  jz4780_gpio_config_pad(gpio_port_f, DDCSDA, Function_alt, 1);

  /* Obtain CPM and I2C objects. */

  cpm = jz4780_cpm_init(cpm_base);

  /* Attempt to set the PCLK source to SCLK_A. */

  jz4780_cpm_set_pclock_source(cpm, 1);
  printf("Peripheral clock: %d\n", jz4780_cpm_get_pclock_frequency(cpm));

  /* Obtain I2C reference. */

  i2c = jz4780_i2c_init(i2c_base, i2c_base_end, cpm, 100000); /* 100 kHz */
  i2c0 = jz4780_i2c_get_channel(i2c, 0);
  i2c4 = jz4780_i2c_get_channel(i2c, 4);

  printf("Scan I2C0...\n");
  i2c_scan(i2c0, irq0cap);

  printf("Scan I2C4...\n");
  i2c_scan(i2c4, irq4cap);

  pmic_out3_control(i2c0, irq0cap);

  ddc_read(i2c4, irq4cap);

  jz4780_i2c_disable(i2c0);
  jz4780_i2c_disable(i2c4);

  /* Detach from the interrupt. */

  err = l4_error(l4_irq_detach(irq0cap)) || l4_error(l4_irq_detach(irq4cap));

  if (err)
    printf("Error detaching from IRQ objects: %ld\n", err);

  printf("Done.\n");

  return 0;
}
